#include <spine/Debug.h>
#include <spine/spine.h>
#include "DxLib.h"

#define SCREEN_W		1280
#define SCREEN_H		720

using namespace spine;

// Spine のデータを纏めた構造体
struct SpineData
{
	SpineData() :
		atlas(NULL),
		skeletonData(NULL),
		stateData(NULL),
		skeleton(NULL),
		state(NULL)
	{
	}

	Atlas *atlas;
	SkeletonData *skeletonData;
	AnimationStateData *stateData;
	Skeleton *skeleton;
	AnimationState *state;
};

// ＤＸライブラリ用テクスチャローダー
class SP_API TextureLoaderDX : public TextureLoader
{
public:
	TextureLoaderDX() {}

	virtual ~TextureLoaderDX() {}

	virtual void load(AtlasPage &page, const String &path);

	virtual void unload(void *texture);
};
void TextureLoaderDX::load(AtlasPage &page, const String &path)
{
	page.setRendererObject( ( void * )LoadGraph( path.buffer() ) );
}
void TextureLoaderDX::unload(void *texture)
{
	DeleteGraph( ( int )texture );
}
TextureLoaderDX *gTextureLoaderDX;

namespace spine {
	SpineExtension *getDefaultExtension() {
		return new DefaultSpineExtension();
	}
}// namespace spine

// バイナリファイルから読み込む関数
void loadBinary(const String &binaryFile, const String &atlasFile, SpineData *spineData) {
	spineData->atlas = new (__FILE__, __LINE__) Atlas(atlasFile, gTextureLoaderDX);
	assert(spineData->atlas != NULL);

	SkeletonBinary binary(spineData->atlas);
	spineData->skeletonData = binary.readSkeletonDataFile(binaryFile);
	assert(spineData->skeletonData);

	spineData->skeleton = new (__FILE__, __LINE__) Skeleton(spineData->skeletonData);
	assert(spineData->skeleton != NULL);

	spineData->stateData = new (__FILE__, __LINE__) AnimationStateData(spineData->skeletonData);
	assert(spineData->stateData != NULL);
	spineData->stateData->setDefaultMix(0.4f);

	spineData->state = new (__FILE__, __LINE__) AnimationState(spineData->stateData);
}

// jsonファイルから読み込む関数
void loadJson(const String &jsonFile, const String &atlasFile, SpineData *spineData) {
	spineData->atlas = new (__FILE__, __LINE__) Atlas(atlasFile, gTextureLoaderDX);
	assert(spineData->atlas != NULL);

	SkeletonJson json(spineData->atlas);
	spineData->skeletonData = json.readSkeletonDataFile(jsonFile);
	assert(spineData->skeletonData);

	spineData->skeleton = new (__FILE__, __LINE__) Skeleton(spineData->skeletonData);
	assert(spineData->skeleton != NULL);

	spineData->stateData = new (__FILE__, __LINE__) AnimationStateData(spineData->skeletonData);
	assert(spineData->stateData != NULL);
	spineData->stateData->setDefaultMix(0.4f);

	spineData->state = new (__FILE__, __LINE__) AnimationState(spineData->stateData);
}

// Spine のデータを解放する関数
void dispose(SpineData *spineData) {
	if( spineData->skeleton     != NULL ){ delete spineData->skeleton;		spineData->skeleton = NULL;		}
	if( spineData->state        != NULL ){ delete spineData->state;			spineData->state = NULL;		}
	if( spineData->stateData    != NULL ){ delete spineData->stateData;		spineData->stateData = NULL;	}
	if( spineData->skeletonData != NULL ){ delete spineData->skeletonData;	spineData->skeletonData = NULL;	}
	if( spineData->atlas        != NULL ){ delete spineData->atlas;			spineData->atlas = NULL;		}
}

// 描画頂点バッファ
int verticesLength;
VERTEX2D *vertices;

// 描画インデックスバッファ
int indicesLength;
unsigned short *indices;

// Spineデータを描画
void drawSkeleton(Skeleton* skeleton) 
{
	// For each slot in the draw order array of the skeleton
	for (size_t i = 0, n = skeleton->getSlots().size(); i < n; ++i) {
		Slot* slot = skeleton->getDrawOrder()[i];

		// Fetch the currently active attachment, continue
		// with the next slot in the draw order if no
		// attachment is active on the slot
		Attachment* attachment = slot->getAttachment();
		if (!attachment) continue;

		// Fetch the blend mode from the slot and
		// translate it to the engine blend mode
		int engineBlendMode;
		switch (slot->getData().getBlendMode()) {
		case BlendMode_Normal:
			engineBlendMode = DX_BLENDMODE_SPINE_NORMAL;
			break;
		case BlendMode_Additive:
			engineBlendMode = DX_BLENDMODE_SPINE_ADDITIVE;
			break;
		case BlendMode_Multiply:
			engineBlendMode = DX_BLENDMODE_SPINE_MULTIPLY;
			break;
		case BlendMode_Screen:
			engineBlendMode = DX_BLENDMODE_SPINE_SCREEN;
			break;
		default:
			// unknown Spine blend mode, fall back to
			// normal blend mode
			engineBlendMode = DX_BLENDMODE_SPINE_NORMAL;
		}

		// Calculate the tinting color based on the skeleton's color
		// and the slot's color. Each color channel is given in the
		// range [0-1], you may have to multiply by 255 and cast to
		// and int if your engine uses integer ranges for color channels.
		Color skeletonColor = skeleton->getColor();
		Color slotColor = slot->getColor();
		Color tint(skeletonColor.r * slotColor.r, skeletonColor.g * slotColor.g, skeletonColor.b * slotColor.b, skeletonColor.a * slotColor.a);

		// Fill the vertices array, indices, and texture depending on the type of attachment
		int texture = NULL;
		if (attachment->getRTTI().isExactly(RegionAttachment::rtti)) {
			// Cast to an spRegionAttachment so we can get the rendererObject
			// and compute the world vertices
			RegionAttachment* regionAttachment = (RegionAttachment*)attachment;

			// Our engine specific Texture is stored in the AtlasRegion which was
			// assigned to the attachment on load. It represents the texture atlas
			// page that contains the image the region attachment is mapped to.
			texture = (int)((AtlasRegion*)regionAttachment->getRendererObject())->page->getRendererObject();

			// Ensure there is enough room for vertices
			// vertices.setSize(4, VERTEX2D());
			if( verticesLength < 4 )
			{
				verticesLength = 4;
				vertices = ( VERTEX2D * )realloc( vertices, sizeof( VERTEX2D ) * verticesLength );
			}
			if( indicesLength < 6 )
			{
				indicesLength = 6;
				indices = ( WORD * )realloc( indices, sizeof( WORD ) * indicesLength );
			}
			indices[ 0 ] = 0;
			indices[ 1 ] = 1;
			indices[ 2 ] = 2;
			indices[ 3 ] = 2;
			indices[ 4 ] = 3;
			indices[ 5 ] = 0;

			// Computed the world vertices positions for the 4 vertices that make up
			// the rectangular region attachment. This assumes the world transform of the
			// bone to which the slot (and hence attachment) is attached has been calculated
			// before rendering via Skeleton::updateWorldTransform(). The vertex positions
			// will be written directoy into the vertices array, with a stride of sizeof(Vertex)
			regionAttachment->computeWorldVertices(slot->getBone(), (float *)vertices, 0, sizeof(VERTEX2D) / sizeof( float ));

			// copy color and UVs to the vertices
			for (size_t j = 0, l = 0; j < 4; j++, l+=2) {
				VERTEX2D& vertex = vertices[j];
				vertex.pos.y = SCREEN_H - vertex.pos.y;
				vertex.dif.r = (BYTE)(tint.r * 255.0f);
				vertex.dif.g = (BYTE)(tint.g * 255.0f);
				vertex.dif.b = (BYTE)(tint.b * 255.0f);
				vertex.dif.a = (BYTE)(tint.a * 255.0f);
				vertex.pos.z = 0.0f;
				vertex.rhw = 1.0f;
				vertex.u = regionAttachment->getUVs()[l];
				vertex.v = regionAttachment->getUVs()[l + 1];
			}

			// Draw the mesh we created for the attachment
			SetDrawBlendMode( engineBlendMode, 255 );
			DrawPolygonIndexed2D(vertices, 4, indices, 2, texture, TRUE);
			SetDrawBlendMode( DX_BLENDMODE_ALPHA, 255 );
		} else if (attachment->getRTTI().isExactly(MeshAttachment::rtti)) {
			// Cast to an MeshAttachment so we can get the rendererObject
			// and compute the world vertices
			MeshAttachment* mesh = (MeshAttachment*)attachment;

			// Ensure there is enough room for vertices
			//         vertices.setSize(mesh->getWorldVerticesLength() / 2, VERTEX2D());
			if( verticesLength < mesh->getWorldVerticesLength() / 2 )
			{
				verticesLength = mesh->getWorldVerticesLength() / 2;
				vertices = ( VERTEX2D * )realloc( vertices, sizeof( VERTEX2D ) * verticesLength );
			}
			if( indicesLength < mesh->getTriangles().size() )
			{
				indicesLength = mesh->getTriangles().size();
				indices = ( WORD * )realloc( indices, sizeof( WORD ) * indicesLength );
			}

			for( int j = 0; j < mesh->getTriangles().size(); j ++ )
			{
				indices[ j ] = mesh->getTriangles()[ j ];
			}

			// Our engine specific Texture is stored in the AtlasRegion which was
			// assigned to the attachment on load. It represents the texture atlas
			// page that contains the image the region attachment is mapped to.
			texture = (int)((AtlasRegion*)mesh->getRendererObject())->page->getRendererObject();

			// Computed the world vertices positions for the vertices that make up
			// the mesh attachment. This assumes the world transform of the
			// bone to which the slot (and hence attachment) is attached has been calculated
			// before rendering via Skeleton::updateWorldTransform(). The vertex positions will
			// be written directly into the vertices array, with a stride of sizeof(Vertex)
			size_t numVertices = mesh->getWorldVerticesLength() / 2;
			mesh->computeWorldVertices(*slot, 0, numVertices * 2, (float *)vertices, 0, sizeof(VERTEX2D) / sizeof( float ));

			// Copy color and UVs to the vertices
			for (size_t j = 0, l = 0; j < numVertices; j++, l+=2) {
				VERTEX2D& vertex = vertices[j];
				vertex.pos.y = SCREEN_H - vertex.pos.y;
				//			vertex.pos.y -= 500.0f;
				vertex.dif.r = (BYTE)(tint.r * 255.0f);
				vertex.dif.g = (BYTE)(tint.g * 255.0f);
				vertex.dif.b = (BYTE)(tint.b * 255.0f);
				vertex.dif.a = (BYTE)(tint.a * 255.0f);
				vertex.pos.z = 0.0f;
				vertex.rhw = 1.0f;
				vertex.u = mesh->getUVs()[l];
				vertex.v = mesh->getUVs()[l + 1];
			}

			// Draw the mesh we created for the attachment
			SetDrawBlendMode( engineBlendMode, 255 );
			DrawPolygonIndexed2D(vertices, numVertices, indices, mesh->getTriangles().size() / 3, texture, TRUE);
			SetDrawBlendMode( DX_BLENDMODE_ALPHA, 255 );
		}
	}
}

int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow )
{
	SpineData spineData;

	// テクスチャローダーを作成しておく
	gTextureLoaderDX = new TextureLoaderDX();

	// ウインドウモードで起動
	ChangeWindowMode( TRUE ) ;

	// 画面モード設定
	SetGraphMode( SCREEN_W, SCREEN_H, 32 ) ;

	// クリアカラーを設定
	SetBackgroundColor(64, 64, 64);

	// ＤＸライブラリの初期化
	if( DxLib_Init() < 0 ) return -1 ;

	DebugExtension debug(SpineExtension::getInstance());
	SpineExtension::setInstance(&debug);

	// Jsonファイルから読み込む
	loadJson(
		"speedy\\export\\speedy-ess.json",
		"speedy\\export\\speedy.atlas",
		&spineData);

	// 描画座標を設定
	spineData.skeleton->setPosition( 1280.0f / 2.0f, 720.0f / 2.0f - 200.0f );

	// アニメーション開始
	spineData.state->setAnimation( 0, spineData.skeletonData->findAnimation( "run" ), true );

	// 描画先を裏画面に変更
	SetDrawScreen( DX_SCREEN_BACK ) ;
	
	// メインループ
	while( ProcessMessage() == 0 )
	{
		ClearDrawScreen();

		// アニメーションを進める
		spineData.state->update( 1.0f / 60.0f );

		// アニメーション結果をスケルトンに反映
		spineData.state->apply( *spineData.skeleton );

		// スケルトンの座標変換状態を更新
		spineData.skeleton->updateWorldTransform();

		// スケルトンを描画
		drawSkeleton( spineData.skeleton );

		// 裏画面の内容を表画面に反映
		ScreenFlip();
	}

	// Spine データを削除
	dispose( &spineData );

	debug.reportLeaks();

	// テクスチャローダーを削除
	delete gTextureLoaderDX;
	gTextureLoaderDX = NULL;

	// 描画頂点データを解放
	if( vertices != NULL )
	{
		free( vertices );
	}

	// 描画インデックスデータを解放
	if( indices != NULL )
	{
		free( indices );
	}

	// ＤＸライブラリの後始末
	DxLib_End() ;

	// ソフトの終了
	return 0 ;
}
